﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Xml;

/// <summary>
/// Age Span class for calculating different number with regards to a persons age
/// </summary>
public class AgeSpan
{
    /// <summary>
    /// The birth date of the person
    /// </summary>
    public DateTime BirthDate { get; set; }

    /// <summary>
    /// The as of date to calculate the AgeSpan for
    /// </summary>
    public DateTime AsOfDate { get; set; }

    /// <summary>
    /// To include the end date inclusively in the calculations.  This will add one extra day to all day totals.
    /// </summary>
    public bool IncludeAsOfDate { get; set; }

    /// <summary>
    /// The number of years from the birth date to the as of date.
    /// </summary>
    public int Years
    {
        get
        {
            int years = this.AsOfDate.Year - this.BirthDate.Year;

            if (this.AsOfDate.Month < this.BirthDate.Month || (this.AsOfDate.Month == this.BirthDate.Month && this.AsOfDate.Day < this.BirthDate.Day))
                --years;

            return years;
        }
    }

    /// <summary>
    /// The number of months from the last full year to the as of month.
    /// </summary>
    public int Months
    {
        get
        {
            int inc = (this.AsOfDate.Day >= this.BirthDate.Day) ? 0 : -1;
            inc = +((this.AsOfDate.Month >= this.BirthDate.Month) ? inc : 12 + inc);
            int months = this.AsOfDate.Month + inc - this.BirthDate.Month;

            return months;
        }
    }

    /// <summary>
    /// The number of days from the last full month to the as of day
    /// </summary>
    /// <remarks>If <see cref="IncludeAsOfDate">IncludeAsOfDate</see> is true one extra day will be added</remarks>
    public int Days
    {
        get
        {

            DateTime prevDate = this.AsOfDate.AddMonths(-1);
            int inc = (this.AsOfDate.Day >= this.BirthDate.Day) ? 0 : -1;
            inc = inc + ((this.IncludeAsOfDate) ? 1 : 0);

            int days = 0;

            if (this.AsOfDate.Day.Equals(this.BirthDate.Day))
            {
                // nothing
            }
            if (this.BirthDate.Day > this.AsOfDate.Day)
            {
                days = DateTime.DaysInMonth(prevDate.Year, prevDate.Month) - this.BirthDate.Day;
                days += this.AsOfDate.Day;
            }
            else
                days = this.AsOfDate.Day - this.BirthDate.Day;

            return (days < 0) ? 0 : days;
        }
    }

    /// <summary>
    /// The total months from the birth date to the as of month
    /// </summary>
    /// <remarks>If the as of day falls before the birth day total months will be less one</remarks>
    public int TotalMonths
    {
        get
        {
            int inc = (this.AsOfDate.Day >= this.BirthDate.Day) ? 0 : -1;
            int months = inc + (((this.AsOfDate.Year - this.BirthDate.Year) * 12) + (this.AsOfDate.Month - this.BirthDate.Month));
            return months;
        }
    }

    /// <summary>
    /// The total days from the birth date to the as of date
    /// </summary>
    /// <remarks>If <see cref="IncludeEndDate">IncludeAsOfDate</see> is true one extra day will be added</remarks>
    public int TotalDays
    {
        get
        {
            return this.AsOfDate.Subtract(this.BirthDate).Days + ((this.IncludeAsOfDate) ? 1 : 0);
        }
    }

    /// <summary>
    /// Creates a AgeSpan instance, AsOfDate set to the current date and IncludeAsOfDate set to false
    /// </summary>
    /// <param name="birthDate">The birth date of the person</param>
    public AgeSpan(DateTime birthDate) : this(birthDate, DateTime.Today) { }

    /// <summary>
    /// Creates a AgeSpan instance, IncludeAsOfDate set to false
    /// </summary>
    /// <param name="birthDate">The birth date of the person</param>
    /// <param name="asOfDate">The as of date for calculations</param>
    public AgeSpan(DateTime birthDate, DateTime asOfDate) : this(birthDate, asOfDate, false) { }

    /// <summary>
    /// Creates a AgeSpan instance
    /// </summary>
    /// <param name="birthDate">The birth date of the person</param>
    /// <param name="asOfDate">The as of date for calculations</param>
    /// <param name="IncludeAsOfDate">To set the calculations to include the AsOfDate inclusively</param>
    public AgeSpan(DateTime birthDate, DateTime asOfDate, bool includeAsOfDate)
    {
        this.BirthDate = new DateTime(birthDate.Year, birthDate.Month, birthDate.Day);
        this.AsOfDate = new DateTime(asOfDate.Year, asOfDate.Month, asOfDate.Day);
        this.IncludeAsOfDate = includeAsOfDate;

        if (this.BirthDate > this.AsOfDate)
        {
            throw new ArgumentOutOfRangeException("birthDate", "birthDate cannot be larger that the asOfDate parameter.");
        }
    }

    /// <summary>
    /// Simple string representing the age span
    /// </summary>
    /// <returns>String</returns>
    public override string ToString()
    {
        return string.Format("years:{0}|months:{1}|days:{2}|total_months:{3}|total_days:{4}", this.Years, this.Months, this.Days, this.TotalMonths, this.TotalDays);
    }

    /// <summary>
    /// Json string representing the age span
    /// </summary>
    /// <returns></returns>
    public string ToJson()
    {
        StringBuilder sb = new StringBuilder();

        sb.Append("{");
        sb.AppendFormat("\"years\":{0},", this.Years);
        sb.AppendFormat("\"months\":{0},", this.Months);
        sb.AppendFormat("\"days\":{0},", this.Days);
        sb.AppendFormat("\"total_months\":{0},", this.TotalMonths);
        sb.AppendFormat("\"total_days\":{0}", this.TotalDays);
        sb.Append("}");

        return sb.ToString();
    }

    /// <summary>
    /// Xml fragement representing the age span
    /// </summary>
    /// <returns></returns>
    public string ToXml()
    {
        using (StringWriter writer = new StringWriter())
        {
            using (XmlWriter xml = new XmlTextWriter(writer))
            {
                xml.WriteStartElement("age_span");
                xml.WriteElementString("years", this.Years.ToString());
                xml.WriteElementString("months", this.Months.ToString());
                xml.WriteElementString("days", this.Days.ToString());
                xml.WriteElementString("total_months", this.TotalMonths.ToString());
                xml.WriteElementString("total_days", this.TotalDays.ToString());
                xml.WriteEndElement();

                return writer.ToString();
            }
        }
    }
}